<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

defined('MOODLE_INTERNAL') || die();

global $CFG;

require_once($CFG->dirroot . '/blocks/zoomonline/classes/ZoomHelper.php');

/**
 * Zoom Online block class.
 *
 * This block provides integration with Zoom, allowing users to easily access and manage Zoom meetings within Moodle.
 *
 * @package    block_zoomonline
 * @copyright  2024 Ciaran Mac Donncha
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class block_zoomonline extends block_base {

    /**
     * Initializes the block's title.
     *
     * This method sets the title for the Zoom Online block, which will be displayed
     * at the top of the block on the Moodle site.
     *
     * @return void
     */
    public function init() {
        global $COURSE;
        $this->title = get_string('pluginname', 'block_zoomonline');
    }

    /**
     * Gets the content for the block.
     *
     * This function is responsible for generating the content displayed within the Zoom Online block.
     * It returns a content object with the text and footer properties that are rendered on the page.
     *
     * @return stdClass|null The content object with properties text and footer, or null if content should not be displayed.
     */
    public function get_content() {
        global $USER, $COURSE, $DB;

        if ($this->content !== null) {
            return $this->content;
        }

        $this->content = new stdClass();
        $this->content->text = '';
        $coursecontext = context_course::instance($COURSE->id);

        $clientid = get_config('block_zoomonline', 'client_id');
        $clientsecret = get_config('block_zoomonline', 'client_secret');
        $accountid = get_config('block_zoomonline', 'account_id');

        if (get_config('block_zoomonline', 'useattendance')) {
            $installedmodules = core_plugin_manager::instance()->get_plugins_of_type('mod');

            if (!array_key_exists('attendance', $installedmodules) &&
                    has_capability('moodle/course:manageactivities', $coursecontext, $USER->id)) {
                $this->content->text =
                        "Warning - Attendance module not installed. Please install <a href='https://moodle.org/plugins/mod_attendance'>mod_attendance</a>.";
                return $this->content;
            }
        }

        if (empty($clientid) || empty($clientsecret) || empty($accountid)) {
            $this->content->text = has_capability('moodle/course:manageactivities', $coursecontext, $USER->id)
                    ? "Warning - Zoom API client_id, client_secret or account_id missing, please check plugin config"
                    : "";
            return $this->content;
        }

        if ($this->coursehasended($COURSE->enddate)) {
            $this->content->text = has_capability('moodle/course:manageactivities', $coursecontext, $USER->id)
                    ? "Warning - This course's end date is in the past, so this will not generate links."
                    : "There are no more online classes for this module.";
            return $this->content;
        }

        $teachers = $this->getcourseteachers();
        $groups = $this->getcoursegroups();
        $this->content->text = $this->generatemeetinghtml($teachers, $groups);

        return $this->content;
    }

    /**
     * Checks if the course has ended.
     *
     * @param int $courseenddate The end date of the course as a Unix timestamp.
     * @return bool Returns true if the course has ended, false otherwise.
     */
    private function coursehasended($courseenddate) {
        return time() > $courseenddate;
    }

    /**
     * Retrieves the teachers for the current course.
     *
     * @return array An array of teacher objects.
     */
    private function getcourseteachers() {
        global $COURSE, $DB;
        $role = $DB->get_record('role', ['shortname' => 'editingteacher']);
        $context = context_course::instance($COURSE->id);
        return get_role_users($role->id, $context);
    }

    /**
     * Retrieves the groups for the current course.
     *
     * @return array An array of group objects, or a default group if none exist.
     */
    private function getcoursegroups() {
        global $COURSE, $DB;
        return $DB->get_records('groups', ['courseid' => $COURSE->id]) ?: [(object) ['id' => 0, 'name' => '']];
    }

    /**
     * Generates HTML for meeting links.
     *
     * @param array $teachers Array of teacher objects.
     * @param array $groups Array of group objects.
     * @return string HTML content for meeting links.
     */
    private function generatemeetinghtml($teachers, $groups) {
        $html = "<div>";
        foreach ($teachers as $teacher) {
            foreach ($groups as $group) {
                $html .= $this->processgroupforteacher($teacher, $group);
            }
        }
        return $html . "</div>" . $this->generatestyleandscript();
    }

    /**
     * Processes a group for a teacher and generates the meeting link.
     *
     * @param object $teacher The teacher object.
     * @param object $group The group object.
     * @return string HTML for the meeting link.
     */
    private function processgroupforteacher($teacher, $group) {
        global $DB, $COURSE, $USER;
        $groupid = $group->id;
        $zoomdata = $DB->get_record('block_zoomonline_links', [
                'course_id' => $COURSE->id,
                'lecturer_id' => $teacher->id,
                'groupid' => $groupid,
        ]);

        $meetingid = $this->managezoommeeting($zoomdata, $teacher, $group);

        if ($meetingid) {
            $this->processzoommeeting($meetingid, $teacher);
        }

        return $meetingid ? $this->generatemeetinglink($teacher, $group, $meetingid, $zoomdata->visible ?? 1) : '';
    }

    /**
     * Manages the Zoom meeting for a teacher and group.
     *
     * @param object|false $zoomdata The Zoom data record, or false if not found.
     * @param object $teacher The teacher object.
     * @param object $group The group object.
     * @return string|null The meeting ID if successful, null otherwise.
     */
    private function managezoommeeting($zoomdata, $teacher, $group) {
        global $DB, $USER;
        $meetingid = $zoomdata->zoom_meeting_id ?? '';
        $hostemail = $teacher->email;

        if ($meetingid && !\block_zoomonline\ZoomHelper::checkmeetingexists($meetingid, $hostemail)) {
            $this->deleteinvalidmeeting($zoomdata);
            $meetingid = '';
        }

        if (!$meetingid && $USER->id === $teacher->id) {
            $meetingid = $this->createzoommeeting($teacher, $group);
        }

        return $meetingid;
    }

    /**
     * Deletes an invalid meeting record from the database.
     *
     * @param object $zoomdata The Zoom data record to delete.
     */
    private function deleteinvalidmeeting($zoomdata) {
        global $DB;
        $DB->delete_records('block_zoomonline', ['id' => $zoomdata->id]);
    }

    /**
     * Creates a new Zoom meeting for a teacher and group.
     *
     * @param object $teacher The teacher object.
     * @param object $group The group object.
     * @return string|null The created meeting ID, or null if creation failed.
     */
    private function createzoommeeting($teacher, $group) {
        global $COURSE, $DB;
        $meetingtopic = $group->id ? "GID_{$group->id}_{$COURSE->idnumber}_{$group->name}" : $COURSE->idnumber;
        $meetingid = \block_zoomonline\ZoomHelper::createmeeting($teacher->email, $meetingtopic);

        if ($meetingid) {
            $zoomdata = (object) [
                    'course_id' => $COURSE->id,
                    'crscode' => $COURSE->idnumber,
                    'lecturer_id' => $teacher->id,
                    'zoom_meeting_id' => $meetingid,
                    'groupid' => $group->id,
            ];
            $DB->insert_record('block_zoomonline_links', $zoomdata);
        }

        return $meetingid;
    }

    /**
     * Processes a Zoom meeting, including recordings and attendance if configured.
     *
     * @param string $meetingid The Zoom meeting ID.
     * @param object $teacher The teacher object.
     */
    private function processzoommeeting($meetingid, $teacher) {
        global $COURSE;
        $this->process_recordings($meetingid, $teacher->firstname, $teacher->lastname);
        if (get_config('block_zoomonline', 'useattendance')) {
            $installedmodules = core_plugin_manager::instance()->get_plugins_of_type('mod');

            if (array_key_exists('attendance', $installedmodules)) {
                $this->process_attendance($meetingid, $COURSE->id);
                $this->check_live_meeting($meetingid, $COURSE->id);
            } else {
                echo "Attendance module is not installed. Skipping attendance processing.";
            }
        }
    }

    /**
     * Process Zoom meeting recordings.
     *
     * This function handles the processing of Zoom recordings for a given meeting.
     * It retrieves the Zoom recordings and processes them based on the provided meeting ID.
     *
     * @param int $meetingid The Zoom meeting ID.
     * @param string|null $firstname Optional. First name of the meeting host.
     * @param string|null $lastname Optional. Last name of the meeting host.
     * @param int|null $crsid Optional. Course ID associated with the Zoom meeting.
     * @param int|null $lecid Optional. Lecturer ID associated with the Zoom meeting.
     *
     * @return void
     */
    public function process_recordings($meetingid, $firstname = '', $lastname = '', $crsid = null, $lecid = null) {
        global $COURSE, $DB;

        $embedtype = get_config('block_zoomonline', 'embedtype');
        $storagetype = get_config('block_zoomonline', 'storagetype');

        if ($lecid) {
            $lecturer = $DB->get_record('user', ['id' => $lecid], 'firstname, lastname');
            $firstname = $lecturer->firstname ?? $firstname;
            $lastname = $lecturer->lastname ?? $lastname;
        }

        $courseid = $crsid ?: $COURSE->id;
        $accesskey = \block_zoomonline\ZoomHelper::getcurrentaccesskey();
        $recordings = \block_zoomonline\ZoomHelper::getmeetingrecordings($meetingid);

        foreach ($recordings as $rec) {
            if ($rec['file_type'] !== 'MP4') {
                continue;
            }

            $videoid = trim($rec['id']);
            $downloadurl = $rec['download_url'];
            $recordingdate = $this->formatrecordingdate($rec['recording_start']);
            $shareurl = $rec['play_url'];
            $vid = trim("$firstname $lastname");

            if ($embedtype == 'book') {
                $scid = $this->createsection();
                $bkid = $this->createbook($vid, $courseid, $scid);

                if ($this->chapterexists($bkid, $recordingdate)) {
                    continue;
                }
            }

            $tempfilepath = \block_zoomonline\ZoomHelper::downloadvideo($downloadurl, $accesskey, $videoid);

            if ($tempfilepath && filesize($tempfilepath) >= 10) {
                if ($storagetype === 'aws') {
                    $s3fileurl = $this->uploadtos3($tempfilepath, $videoid);
                    unlink($tempfilepath);
                } else if ($storagetype === 'local') {
                    $localfileurl = $this->savetolocalfolder($tempfilepath, $videoid);
                }

                if (isset($s3fileurl) || isset($localfileurl)) {
                    \block_zoomonline\ZoomHelper::deletezoomrecording($meetingid, $videoid);
                    $fileurl = $storagetype === 'aws' ? $s3fileurl : $localfileurl;
                    if ($embedtype == 'book') {
                        $bkid = $this->addchaptertobook($bkid, $fileurl, $shareurl, $firstname, $lastname,
                                $recordingdate, $vid);
                    } else if ($embedtype === 'label') {
                        $this->addvideotomoodle($courseid, $videoid, $fileurl, $firstname, $lastname,
                                $recordingdate);
                    }
                }
            }
        }
    }

    /**
     * Formats the recording date.
     *
     * Converts the given string time into a formatted date string in the 'd/m/Y H:i' format.
     * It adjusts the time based on the server's default timezone.
     *
     * @param string $stringtime The original string representing the recording time.
     *
     * @return string The formatted date in 'd/m/Y H:i' format.
     */
    private function formatrecordingdate($stringtime) {
        $timestamp = strtotime($stringtime);
        $date = new DateTime();
        $date->setTimestamp($timestamp);
        $date->setTimezone(new DateTimeZone(date_default_timezone_get()));
        return $date->format('d/m/Y H:i');
    }

    /**
     * Creates or retrieves the 'ONLINE RECORDINGS' section in the course.
     *
     * This method checks if a section named 'ONLINE RECORDINGS' exists in the current course.
     * If it doesn't exist, it creates a new section at the end of the course.
     * If it already exists, it retrieves the section ID.
     *
     * @return int The section ID of the 'ONLINE RECORDINGS' section.
     */
    private function createsection() {
        global $COURSE, $DB, $USER;

        $cid = $COURSE->id;
        $sectionid = 0;
        $sqlsectionexists =
                "SELECT course,section FROM {course_sections} where course = $cid AND name = 'ONLINE RECORDINGS' limit 1";
        $sectionexists = $DB->get_records_sql($sqlsectionexists);

        if (!$sectionexists) {
            $recsection = new stdClass();
            $recsection->course = $COURSE->id;
            $sqlsectop =
                    "SELECT course,section FROM {course_sections} where course = $cid order by section desc limit 1";
            $sectop = $DB->get_records_sql($sqlsectop);
            $sectionid = $sectop[$cid]->section + 1;
            $recsection->section = $sectionid;
            $recsection->name = 'ONLINE RECORDINGS';
            $recsection->summaryformat = FORMAT_HTML;
            $recsection->sequence = '';

            $id = $DB->insert_record("course_sections", $recsection);
        } else {
            $sectionid = $sectionexists[$cid]->section;
        }
        return $sectionid;
    }

    /**
     * Creates a new book module or retrieves an existing one for storing recordings.
     *
     * This method checks if a book module with a name starting with the given video ID exists in the course.
     * If it doesn't exist, it creates a new book module in the specified section.
     * If it already exists, it returns the existing book's ID.
     *
     * @param string $vid The video ID, used as the base for the book name.
     * @param int $cid The course ID where the book will be created.
     * @param int $sectionid The section ID where the book will be placed.
     * @return int|null The ID of the created or existing book module, or null if an error occurred.
     * @throws Exception If there's an error during book creation or retrieval.
     */
    private function createbook($vid, $cid, $sectionid) {
        global $CFG, $COURSE, $DB;

        try {
            $origname = $vid;

            $vid = $DB->sql_like_escape($vid); // Escape special characters for LIKE clause .

            $bookexists =
                    $DB->get_record_sql("SELECT id FROM {book} WHERE course = ? AND name LIKE ?", [$cid, "$vid%"]);

            if (!$bookexists) {
                $book = new stdClass();
                $book->course = $cid;
                $book->visible = 1;
                $book->section = $sectionid;
                $book->module = $DB->get_field('modules', 'id', ['name' => 'book']);
                $book->modulename = 'book';
                $book->intro = "Recordings for '$COURSE->shortname'";
                $book->introformat = FORMAT_HTML;
                $book->name = $origname;

                require_once($CFG->dirroot . '/mod/label/lib.php');
                require_once($CFG->dirroot . '/course/modlib.php');

                $moduleinstance = add_moduleinfo($book, $COURSE);
                return $moduleinstance->instance;
            } else {
                return $bookexists->id;
            }
        } catch (Exception $ex) {
            return null;
        }
    }

    /**
     * Checks if a chapter with the given recording date already exists in the specified book.
     *
     * @param int $bookid The ID of the book to check.
     * @param string $recordingdate The date of the recording, used as the chapter title.
     * @return bool True if the chapter exists, false otherwise.
     */
    private function chapterexists($bookid, $recordingdate) {
        global $DB;
        $t = $DB->record_exists('book_chapters', ['bookid' => $bookid, 'title' => $recordingdate]);

        return $t;
    }

    /**
     * Uploads a file to Amazon S3 storage.
     *
     * This method uses the AWS S3 SDK to upload a file to the configured S3 bucket.
     * It retrieves AWS credentials and bucket information from the plugin configuration.
     *
     * @param string $filepath The local path of the file to be uploaded.
     * @param string $filename The name to be given to the file in S3 (without .mp4 extension).
     * @return string|null The URL of the uploaded file in S3 if successful, null otherwise.
     * @throws Exception If there's an error during the upload process.
     */
    private function uploadtos3($filepath, $filename) {
        global $CFG;

        $filename .= '.mp4';

        $awskey = get_config('block_zoomonline', 'aws_key');
        $awssecret = get_config('block_zoomonline', 'aws_secret');
        $bucket = get_config('block_zoomonline', 'aws_bucket');
        $region = get_config('block_zoomonline', 'aws_region');

        try {
            if (!class_exists('S3')) {
                require_once($CFG->dirroot . '/repository/s3/S3.php');
            }

            S3::setAuth($awskey, $awssecret);

            if (S3::putObject(S3::inputFile($filepath), $bucket, $filename, S3::ACL_PUBLIC_READ)) {
                return "https://$bucket.s3.$region.amazonaws.com/$filename";
            } else {
                debugging("S3 Upload failed: Unable to upload $filename to bucket $bucket.", DEBUG_DEVELOPER);

                return null;
            }

        } catch (Exception $e) {

            debugging('S3 Upload failed: ' . $e->getMessage(), DEBUG_DEVELOPER);
            return null;
        }
    }

    /**
     * Saves a video file to the local Moodle file system.
     *
     * This method handles the storage of a video file in Moodle's file system,
     * creating a new file or returning the URL of an existing file
     *
     * @param string $tempfilepath The temporary path of the file to be saved.
     * @param string $videoid The unique identifier for the video.
     * @return string The URL of the saved or existing file.
     * @throws moodle_exception If there's an error in the block instance, context, or file saving process.
     * @throws coding_exception If the context is invalid or not set to the block level.
     */
    private function savetolocalfolder($tempfilepath, $videoid) {
        global $CFG;

        if (!isset($this->instance->id) || !$this->instance->id) {
            throw new moodle_exception('Block instance ID is not set or invalid.');
        }

        try {
            $context = context_block::instance($this->instance->id);
        } catch (Exception $e) {
            throw new moodle_exception('Invalid context ID specified for block_zoomonline.');
        }

        if (!$context || $context->contextlevel != CONTEXT_BLOCK) {
            throw new coding_exception('Context is not valid or not set to the block level.');
        }

        $fs = get_file_storage();

        $filerecord = [
                'contextid' => $context->id,        // Use the retrieved block context.
                'component' => 'block_zoomonline',  // Component name for your block.
                'filearea' => 'zoom_videos',        // Custom file area name for organization.
                'itemid' => 0,                      // Item ID (use 0 if not needed).
                'filepath' => '/',                  // Filepath (always start with '/').
                'filename' => $videoid . '.mp4',     // The name of the file.
        ];

        $existingfile = $fs->get_file(
                $filerecord['contextid'],
                $filerecord['component'],
                $filerecord['filearea'],
                $filerecord['itemid'],
                $filerecord['filepath'],
                $filerecord['filename']
        );

        if ($existingfile) {
            return moodle_url::make_pluginfile_url(
                    $existingfile->get_contextid(),
                    $existingfile->get_component(),
                    $existingfile->get_filearea(),
                    $existingfile->get_itemid(),
                    $existingfile->get_filepath(),
                    $existingfile->get_filename()
            )->out(false);
        }

        $storedfile = $fs->create_file_from_pathname($filerecord, $tempfilepath);

        if ($storedfile) {
            return moodle_url::make_pluginfile_url(
                    $context->id,
                    'block_zoomonline',
                    'zoom_videos',
                    0,
                    '/',
                    $videoid . '.mp4'
            )->out(false);
        } else {
            throw new moodle_exception('Error saving file to Moodle data.');
        }
    }

    /**
     * Adds a new chapter to a book module with video content.
     *
     * This method creates a new chapter in the specified book, including
     * a video player, recording information, and a link to view the video in Zoom.
     *
     * @param int $bookid The ID of the book to add the chapter to.
     * @param string $s3fileurl The URL of the video file in S3.
     * @param string $shareurl The Zoom share URL for the video.
     * @param string $firstname The first name of the recorder.
     * @param string $lastname The last name of the recorder.
     * @param string $recordingdate The date of the recording.
     * @param string $vid The video ID.
     */
    private function addchaptertobook($bookid, $s3fileurl, $shareurl, $firstname, $lastname, $recordingdate, $vid) {
        global $DB;

        $content = "<video width='300' height='175' oncontextmenu='return false;' controls='controls' controlsList='nodownload'>
                    <source src='$s3fileurl' type='video/mp4'>$vid
                </video>
                <p>$firstname $lastname - $recordingdate</p>
                <p><a target='_blank' href='$shareurl'>Click here to play video in Zoom with transcript, you must be signed into Zoom</a></p>";

        $chapter = (object) [
                'bookid' => $bookid,
                'content' => $content,
                'title' => $recordingdate,
                'hidden' => 1,
        ];
        $DB->insert_record('book_chapters', $chapter);
    }

    /**
     * Processes attendance for a Zoom meeting.
     *
     * This method retrieves meeting data, checks for existing attendance records,
     * and processes attendance data for participants if it hasn't been done before.
     *
     * @param string $meetingid The Zoom meeting ID.
     * @param int $courseid The Moodle course ID.
     * @return bool False if the meeting data is invalid or attendance has already been processed, true otherwise.
     */
    public function process_attendance($meetingid, $courseid) {
        $meetingdata = \block_zoomonline\ZoomHelper::getmeetingdata($meetingid);

        if (is_array($meetingdata) && isset($meetingdata['code']) && $meetingdata['code'] === 3301) {
            return false;
        }

        if (!$meetingdata) {
            return false;
        }
        $meetingstartwtz = $meetingdata['start_time'];
        $meetingduration = $meetingdata['duration'];
        $topic = $meetingdata['topic'];
        $meetingstartec = $this->getroundedmeetingstart($meetingstartwtz);
        $meetingstart = date('Y/m/d h:i:s', $meetingstartec);

        if (!$this->attendanceexists($meetingid, $meetingstart) && $meetingstartec) {
            $attendees = $this->getallmeetingparticipants($meetingid, $meetingduration);
            if (empty($attendees)) {
                return false;
            }

            $consolidatedattendees = $this->consolidateattendancedata($attendees);
            $this->processfinalattendance($consolidatedattendees, $meetingstartwtz, $meetingduration, $courseid,
                    $topic);

            $this->markattendanceprocessed($meetingid, $meetingstart);
        } else {
            return false;
        }
    }

    /**
     * Get the rounded meeting start time.
     *
     * @param string $meetingstartwtz
     * @return int Rounded meeting start timestamp.
     */
    private function getroundedmeetingstart($meetingstartwtz) {
        $meetingstarte = strtotime($meetingstartwtz);
        return $meetingstarte - ($meetingstarte % 60); // Round to nearest minute.
    }

    /**
     * Check if attendance already exists for a meeting.
     *
     * @param string $meetingid
     * @param string $meetingstart
     * @return bool
     */
    private function attendanceexists($meetingid, $meetingstart) {
        global $DB;

        return $DB->record_exists('block_zoomonline_att_check',
                ['zoom_meeting_id' => $meetingid, 'start_date' => $meetingstart]);
    }

    /**
     * Get all meeting participants from Zoom API.
     *
     * @param string $meetingid
     * @param int $meetingduration
     * @return array List of participants.
     */
    private function getallmeetingparticipants($meetingid, $meetingduration) {
        $attendees = [];
        $nextpagetoken = "";

        while (true) {
            $data = \block_zoomonline\ZoomHelper::getparticipants($meetingid, $nextpagetoken);

            if (!$data) {
                break;
            }
            if (isset($data['participants']) && $meetingduration >= 5) {
                foreach ($data['participants'] as $USER) {
                    $attuser = $this->createattuserobject($USER);
                    if ($attuser) {
                        $attendees[] = $attuser;
                    }
                }
            }

            if (empty($data['next_page_token'])) {
                break;
            } else {
                $nextpagetoken = $data['next_page_token'];
            }
        }

        return $attendees;
    }

    /**
     * Create an attendance user object from Zoom participant data.
     *
     * @param array $USER
     * @return stdClass|null
     */
    private function createattuserobject($USER) {
        $stuid = filter_var($USER['name'], FILTER_SANITIZE_NUMBER_INT);
        if ($stuid > 0) {
            $attuser = new stdClass();
            $attuser->name = $stuid;
            $attuser->join_time = $USER['join_time'];
            $attuser->leave_time = $USER['leave_time'];
            $attuser->duration = $USER['duration'];
            $attuser->email = $USER['user_email'];
            return $attuser;
        }
        return null;
    }

    /**
     * Consolidate multiple attendance records into a single one.
     *
     * @param array $attendees
     * @return array Consolidated attendees.
     */
    private function consolidateattendancedata($attendees) {
        $consolidated = [];

        foreach ($attendees as $att) {
            $nametofind = $att->name;
            if ($nametofind) {
                $existing = $this->findinarray($consolidated, 'name', $nametofind);

                if (!$existing) {
                    $groupedattendees = array_filter($attendees, function($attone) use ($nametofind) {
                        return $attone->name === $nametofind;
                    });

                    usort($groupedattendees, function($a, $b) {
                        return strtotime($a->join_time) <=> strtotime($b->join_time);
                    });

                    $attstud = new stdClass();
                    $attstud->name = $nametofind;
                    $attstud->join_time = end($groupedattendees)->join_time;
                    $attstud->leave_time = reset($groupedattendees)->leave_time;
                    $attstud->duration = array_reduce($groupedattendees, function($carry, $item) {
                        return $carry + $item->duration;
                    }, 0);

                    $consolidated[] = $attstud;
                }
            }
        }

        return $consolidated;
    }

    /**
     * Find an element in an array by a key-value pair.
     *
     * @param array $array
     * @param string $key
     * @param mixed $value
     * @return mixed|null
     */
    private function findinarray($array, $key, $value) {
        foreach ($array as $element) {
            if ($element->$key === $value) {
                return $element;
            }
        }
        return null;
    }

    /**
     * Process final attendance data and update logs.
     *
     * @param array $attendees
     * @param string $meetingstartwtz
     * @param int $meetingduration
     * @param int $courseid
     * @param string $topic
     */
    private function processfinalattendance($attendees, $meetingstartwtz, $meetingduration, $courseid, $topic) {
        $length = count($attendees);
        $x = 1;

        foreach ($attendees as $attendee) {
            $close = ($x === $length) ? 1 : 0;
            $this->updateatt($attendee, $meetingstartwtz, $meetingduration, $courseid, $close, $topic);
            $x++;
        }
    }

    /**
     * Updates attendance records for a student in a Zoom meeting.
     *
     * This method calculates the student's attendance status, retrieves or creates
     * the necessary attendance records, and updates or creates attendance sessions
     * as needed.
     *
     * @param object $finstuatt The student's attendance data.
     * @param string $meetingstart The start time of the meeting.
     * @param int $meetingduration The duration of the meeting in minutes.
     * @param int $zcrsid The Zoom course ID.
     * @param bool $close Whether to close the attendance session.
     * @param string $topic The topic or name of the meeting.
     */
    private function updateatt($finstuatt, $meetingstart, $meetingduration, $zcrsid, $close, $topic) {
        global $CFG;

        $status = $this->calculatestudentstatus($finstuatt, $meetingstart, $meetingduration);
        $studentid = $this->getstudentidbyusername($finstuatt->name);

        if (!$studentid) {
            return;
        }

        $attendancerecord = $this->getattendancerecord($zcrsid);

        if ($attendancerecord) {
            $sessionexists =
                    $this->processattendancesession($attendancerecord, $meetingstart, $meetingduration, $status,
                            $zcrsid, $studentid, $close);

            if (!$sessionexists) {
                $this->createattsess($attendancerecord->id, $this->getroundedmeetingstarttime($meetingstart),
                        $meetingduration * 60, $status, $zcrsid, $studentid, $topic, false);
            }
        } else {
            $this->createnewattendanceinstance($zcrsid, $meetingstart, $meetingduration, $status, $studentid, $topic);
        }
    }

    /**
     * Calculate the student's status based on their attendance and meeting details.
     *
     * @param object $finstuatt
     * @param string $meetingstart
     * @param int $meetingduration
     * @return string The student's attendance status (P, L).
     */
    private function calculatestudentstatus($finstuatt, $meetingstart, $meetingduration) {
        $status = "P";
        $latetime = get_config('block_zoomonline', 'latetime');

        if ($latetime == 0) {
            $latetime = 900;
        }

        $attendancethreshold = get_config('block_zoomonline', 'attendance_threshold');

        if ($attendancethreshold == 0) {
            $attendancethreshold = 70;
        }

        $earlydeparturethreshold = get_config('block_zoomonline', 'early_departure_threshold');

        if ($earlydeparturethreshold == 0) {
            $earlydeparturethreshold = 900;
        }
        $meetingstarttime = strtotime($meetingstart);
        $meetingstartec = $meetingstarttime - ($meetingstarttime % 60); // Round to nearest minute.
        $studentjointime = strtotime($finstuatt->join_time);

        $studentlate = max(0, $studentjointime - $meetingstartec);
        $duration = $meetingduration * 60;
        $attendedpercentage = ($finstuatt->duration * 100) / $duration;

        if ($attendedpercentage < $attendancethreshold || $studentlate > $latetime) {
            $status = "L"; // Late
        }

        if ($duration > 1800) {
            $sessionendtime = $meetingstartec + $duration;
            $studentendtime = $studentjointime + $finstuatt->duration;

            if (($sessionendtime - $studentendtime) > $earlydeparturethreshold) {
                $status = "L"; // Late if student leaves too early.
            }
        }

        return $status;
    }

    /**
     * Get the student's ID by their username.
     *
     * @param string $username
     * @return int|null The student's ID or null if not found.
     */
    private function getstudentidbyusername($username) {
        global $DB;
        $student = $DB->get_record('user', ['username' => $username]);
        return $student ? $student->id : null;
    }

    /**
     * Retrieve the attendance record for the course.
     *
     * @param int $zcrsid The course ID.
     * @return object|null The attendance record or null if not found.
     */
    private function getattendancerecord($zcrsid) {
        global $DB;
        return $DB->get_record('attendance', ['course' => $zcrsid]);
    }

    /**
     * Process the attendance session, updating or creating a log as necessary.
     *
     * @param object $attendancerecord
     * @param string $meetingstart
     * @param int $meetingduration
     * @param string $status
     * @param int $zcrsid
     * @param int $studentid
     * @param int $close
     * @return bool True if session exists, false otherwise.
     */
    private function processattendancesession($attendancerecord, $meetingstart, $meetingduration, $status, $zcrsid,
            $studentid, $close) {

        global $DB;
        $meetingstarttime = $this->getroundedmeetingstarttime($meetingstart);

        $duration = $meetingduration * 60;
        $attsesssql =
                "SELECT attendanceid, id, duration FROM {attendance_sessions} WHERE attendanceid = ? AND sessdate = ? AND (duration = ? OR duration = '3600')";
        $attsess = $DB->get_records_sql($attsesssql, [$attendancerecord->id, $meetingstarttime, $duration]);

        if ($attsess) {
            $attsessid = reset($attsess)->id;
            if (reset($attsess)->duration == '3600' && $duration != 3600) {
                $DB->execute("UPDATE {attendance_sessions} SET duration = ? WHERE attendanceid = ? AND sessdate = ? AND duration = '3600'",
                        [$duration, $attendancerecord->id, $meetingstarttime]);
            }

            $stuattsesssql =
                    "SELECT sessionid, studentid, statusid, remarks, id FROM {attendance_log} WHERE sessionid = ? AND studentid = ?";

            $stuattsess = $DB->get_records_sql($stuattsesssql, [$attsessid, $studentid]);
            $attstatus = $DB->get_record('attendance_statuses',
                    ['acronym' => $status, 'attendanceid' => $attendancerecord->id]);

            if (!$stuattsess || $close == 1) {
                $this->createattlog($attsessid, $attstatus->id, $studentid, $meetingstarttime, $zcrsid);
            } else {
                $this->updateattendancelogifneeded($stuattsess[$attsessid], $attstatus->id, $meetingstarttime);
            }

            return true;
        }

        return false;
    }

    /**
     * Get the rounded meeting start time.
     *
     * @param string $starttime
     * @return int Rounded meeting start timestamp.
     */
    private function getroundedmeetingstarttime($starttime) {
        $meetingstarte = strtotime($starttime);
        return $meetingstarte - ($meetingstarte % 60); // Round to nearest minute.
    }

    /**
     * Creates an attendance log entry for a student.
     *
     * This method checks if an attendance log already exists for the given session and student.
     * If no log exists, it creates a new attendance log entry with the provided details.
     *
     * @param int $sessid The ID of the attendance session.
     * @param int $attstatus The status ID representing the student's attendance (e.g., present, absent).
     * @param int $studentid The ID of the student.
     * @param int $sessdate The timestamp of when the attendance was taken.
     * @param int $zcrsid The Zoom course ID (not used in the method, consider removing if unnecessary).
     */
    private function createattlog($sessid, $attstatus, $studentid, $sessdate, $zcrsid) {
        global $DB;

        $existinglog = $DB->record_exists('attendance_log', [
                'sessionid' => $sessid,
                'studentid' => $studentid,
        ]);

        if (!$existinglog) {
            $newattdata = (object) [
                    'sessionid' => $sessid,
                    'studentid' => $studentid,
                    'statusid' => $attstatus,
                    'statusset' => '9,11,12,10',
                    'remarks' => '',
                    'timetaken' => $sessdate,
                    'takenby' => '2',
            ];
            $DB->insert_record('attendance_log', $newattdata);
        }
    }

    /**
     * Update the attendance log if the system previously auto-recorded it.
     *
     * @param object $attendancelog
     * @param int $attstatusid
     * @param int $meetingstarttime
     */
    private function updateattendancelogifneeded($attendancelog, $attstatusid, $meetingstarttime) {
        global $DB;

        if ($attendancelog->remarks == 'system auto recorded') {
            $attendancelog->timetaken = $meetingstarttime;
            $attendancelog->statusid = $attstatusid;
            $attendancelog->remarks = 'Zoom';
            $DB->update_record('attendance_log', $attendancelog);
        }
    }

    /**
     * Creates a new attendance session and optionally logs student attendance.
     *
     * This method generates a new attendance session with the provided details,
     * assigns a group if applicable, ensures attendance statuses exist,
     * and optionally creates an attendance log for a student.
     *
     * @param int $attinstanceid The ID of the attendance instance.
     * @param int $sessdate The timestamp for the session date.
     * @param int $duration The duration of the session in seconds.
     * @param string $status The attendance status acronym (e.g., 'P' for present).
     * @param int $zcrsid The Zoom course ID.
     * @param int $studentid The ID of the student to log attendance for.
     * @param string $topic The topic or description of the session.
     * @param bool $pre Whether this is a pre-session creation (default false).
     */
    private function createattsess($attinstanceid, $sessdate, $duration, $status, $zcrsid, $studentid, $topic, $pre = false) {
        global $DB;

        $password = $this->generaterandompassword(8);

        $attdata = (object) [
                'attendanceid' => $attinstanceid,
                'sessdate' => $sessdate,
                'duration' => $duration,
                'lasttaken' => $sessdate,
                'lasttakenby' => 2,
                'timemodified' => $sessdate,
                'descriptionformat' => 1,
                'calendarevent' => 0,
                'studentscanmark' => 1,
                'studentpassword' => $password,
                'includeqrcode' => 0,
                'rotateqrcode' => 1,
                'automark' => 2,
                'description' => $topic,
        ];

        $this->assigngroupifexists($attdata, $topic, $zcrsid);

        $sessionid = $DB->insert_record('attendance_sessions', $attdata);

        if ($sessionid) {
            $this->ensureattendancestatuses($attinstanceid, $status);

            if (!$pre) {
                $attstatus = $DB->get_record('attendance_statuses',
                        ['acronym' => $status, 'attendanceid' => $attinstanceid]);
                $this->createattlog($sessionid, $attstatus->id, $studentid, $sessdate, $zcrsid);
            }
        }
    }

    /**
     * Generate a random password of the specified length.
     *
     * @param int $length The length of the password.
     * @return string The generated password.
     */
    private function generaterandompassword($length) {
        $chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
        return substr(str_shuffle($chars), 0, $length);
    }

    /**
     * Assign group to attendance session if topic contains group information.
     *
     * @param stdClass $attdata The attendance session data object.
     * @param string $topic The topic which may contain group information.
     * @param int $zcrsid The course ID.
     */
    private function assigngroupifexists(&$attdata, $topic, $zcrsid) {
        global $DB;

        $namepieces = explode("_", $topic);

        if (count($namepieces) > 1 && $namepieces[0] === "GID") {
            $groupidnumber = $namepieces[1];
            $group = $DB->get_record_sql("SELECT id FROM {groups} WHERE id = ? AND courseid = ?",
                    [$groupidnumber, $zcrsid]);

            if ($group) {
                $attdata->groupid = $group->id;
            }
        }
    }

    /**
     * Ensure the required attendance statuses exist for the given instance.
     *
     * @param int $attinstanceid The attendance instance ID.
     * @param string $status The status acronym.
     */
    private function ensureattendancestatuses($attinstanceid, $status) {
        global $DB;

        $existingstatus = $DB->get_record('attendance_statuses', ['acronym' => $status, 'attendanceid' => $attinstanceid]);

        if (!$existingstatus) {
            $this->createssstatus($attinstanceid, "P", "Present", "2.00");
            $this->createssstatus($attinstanceid, "A", "Absent", "0.00");
            $this->createssstatus($attinstanceid, "L", "Late", "0.00");
            $this->createssstatus($attinstanceid, "E", "Excused", "2.00");
        }
    }

    /**
     * Creates a new attendance status entry for a given attendance instance.
     *
     * This method adds a new status to the attendance_statuses table with the specified
     * acronym, description, and grade. The 'setunmarked' field is automatically set to 1
     * if the acronym is 'A', indicating an absent status.
     *
     * @param int $attinstanceid The ID of the attendance instance to which the status belongs.
     * @param string $acronym The acronym representing the attendance status (e.g., 'P' for present, 'A' for absent).
     * @param string $desc A description of the attendance status.
     * @param float $grade The grade value associated with this attendance status.
     */
    private function createssstatus($attinstanceid, $acronym, $desc, $grade) {
        global $DB;

        $attsesse = (object) [
                'attendanceid' => $attinstanceid,
                'acronym' => $acronym,
                'description' => $desc,
                'grade' => $grade,
                'visible' => 1,
                'deleted' => 0,
                'setnumber' => 0,
                'setunmarked' => ($acronym === 'A') ? 1 : 0, // Automatically set 'setunmarked' for 'A' acronym.
        ];

        $DB->insert_record('attendance_statuses', $attsesse);
    }

    /**
     * Create a new attendance instance and session for the course.
     *
     * @param int $zcrsid
     * @param string $meetingstart
     * @param int $meetingduration
     * @param string $status
     * @param int $studentid
     * @param string $topic
     */
    private function createnewattendanceinstance($zcrsid, $meetingstart, $meetingduration, $status, $studentid, $topic) {
        global $CFG, $DB;

        require_once($CFG->dirroot . '/mod/attendance/lib.php');
        require_once($CFG->dirroot . '/course/modlib.php');

        $attendance = (object) [
                'course' => $zcrsid,
                'name' => 'Attendance',
                'grade' => 100,
                'timemodified' => time(),
                'intro' => '',
                'introformat' => 1,
                'subnet' => '',
                'sessiondetailspos' => 'left',
                'showsessiondetails' => 1,
                'showextrauserdetails' => 1,
        ];
        $attinstanceid = $DB->insert_record('attendance', $attendance);

        $moduleid = $DB->get_field('modules', 'id', ['name' => 'attendance']);
        $coursemodule = (object) [
                'course' => $zcrsid,
                'module' => $moduleid,
                'instance' => $attinstanceid,
                'section' => 0,
                'added' => time(),
        ];
        $cmid = $DB->insert_record('course_modules', $coursemodule);
        course_add_cm_to_section($zcrsid, $cmid, 3);
        rebuild_course_cache($zcrsid, true);

        $this->createattsess($attinstanceid, $this->getroundedmeetingstarttime($meetingstart), $meetingduration * 60,
                $status, $zcrsid, $studentid, $topic, false);
    }

    /**
     * Mark the attendance as processed for a meeting.
     *
     * @param string $meetingid
     * @param string $meetingstart
     */
    private function markattendanceprocessed($meetingid, $meetingstart) {
        global $DB;

        $record = (object) [
                'zoom_meeting_id' => $meetingid,
                'start_date' => $meetingstart,
        ];
        $DB->insert_record("block_zoomonline_att_check", $record);
    }

    /**
     * Checks the status of a live Zoom meeting and manages attendance sessions.
     *
     * This method verifies if a Zoom meeting has started and retrieves its metrics.
     * It then calculates the rounded start time and checks for an existing attendance
     * instance in the course. If an attendance instance exists, it ensures a session
     * is created or updated. If not, it creates a new attendance module and session.
     *
     * @param string $meetingid The ID of the Zoom meeting to check.
     * @param int $courseid The ID of the Moodle course associated with the meeting.
     */
    public function check_live_meeting($meetingid, $courseid) {
        $meetingdata = \block_zoomonline\ZoomHelper::getmeetingstatus($meetingid);

        if ($meetingdata['status'] !== "started") {
            return;
        }

        $meetingmetrics = \block_zoomonline\ZoomHelper::getmeetingmetrics($meetingid);

        $meetingstarttime = $this->getroundedmeetingstarttime($meetingmetrics['start_time']);

        if (!$meetingstarttime) {
            return;
        }

        $attinstanceid = $this->getattendanceinstance($courseid);

        if ($attinstanceid) {
            $this->checkorcreatesession($attinstanceid, $meetingstarttime, $courseid,
                    $meetingmetrics['topic']);
        } else {
            $attinstanceid = $this->createnewattendancemodule($courseid);
            $this->checkorcreatesession($attinstanceid, $meetingstarttime, $courseid,
                    $meetingmetrics['topic']);
        }
    }

    /**
     * Get the attendance instance for the course.
     *
     * @param int $courseid
     * @return int|null The attendance instance ID or null if not found.
     */
    private function getattendanceinstance($courseid) {
        global $DB;
        $attendancerecord = $DB->get_record('attendance', ['course' => $courseid]);
        return $attendancerecord ? $attendancerecord->id : null;
    }

    /**
     * Check if an attendance session exists for a given time, and create it if not.
     *
     * @param int $attinstanceid
     * @param int $meetingstarttime
     * @param int $courseid
     * @param string $topic
     */
    private function checkorcreatesession($attinstanceid, $meetingstarttime, $courseid, $topic) {
        global $DB;

        $attsessexists = $DB->record_exists('attendance_sessions', [
                'attendanceid' => $attinstanceid,
                'sessdate' => $meetingstarttime,
        ]);

        if (!$attsessexists) {
            $status = "P";
            $this->createattsess($attinstanceid, $meetingstarttime, 3600, $status, $courseid, "None", $topic, true);
        }
    }

    /**
     * Create a new attendance module for the course.
     *
     * @param int $courseid
     * @return int The created attendance instance ID.
     */
    private function createnewattendancemodule($courseid) {

        global $CFG, $DB;

        require_once($CFG->dirroot . '/mod/attendance/lib.php');
        require_once($CFG->dirroot . '/course/modlib.php');

        $attendance = (object) [
                'course' => $courseid,
                'name' => 'Attendance',
                'grade' => 100,
                'timemodified' => time(),
                'intro' => '',
                'introformat' => 1,
                'subnet' => '',
                'sessiondetailspos' => 'left',
                'showsessiondetails' => 1,
                'showextrauserdetails' => 1,
        ];
        $attinstanceid = $DB->insert_record('attendance', $attendance);

        $moduleid = $DB->get_field('modules', 'id', ['name' => 'attendance']);
        $coursemodule = (object) [
                'course' => $courseid,
                'module' => $moduleid,
                'instance' => $attinstanceid,
                'section' => 0,
                'added' => time(),
        ];
        $cmid = $DB->insert_record('course_modules', $coursemodule);

        course_add_cm_to_section($courseid, $cmid, 0);
        rebuild_course_cache($courseid, true);

        return $attinstanceid;
    }

    /**
     * Generates an HTML link for joining a Zoom meeting.
     *
     * This method creates a formatted HTML snippet containing information about
     * the meeting and a link to join it. The appearance and functionality of the
     * link vary based on the user's access rights and the meeting's visibility.
     *
     * @param object $teacher An object containing the teacher's first name and last name.
     * @param object $group An object containing the group's name.
     * @param string $meetingid The Zoom meeting ID.
     * @param int $visibility The visibility status of the meeting (2 for hidden, other values for visible).
     * @return string HTML snippet for the meeting link, or an empty string if the user doesn't have access.
     */
    private function generatemeetinglink($teacher, $group, $meetingid, $visibility) {
        $linkroot = "https://@zoom.us/j/";
        $isvisible = $visibility != 2;
        $userhasaccess = $this->userhasaccess($teacher);

        if ($userhasaccess) {
            $eyeicon = $isvisible
                    ?
                    "<span class='live_join_class_eye' data-id='{$meetingid}' data-visible='2'><i class='icon fa fa-eye fa-fw' aria-hidden='true'></i></span>"
                    :
                    "<span class='live_join_class_eye' data-id='{$meetingid}' data-visible='1'><i class='icon fa fa-eye-slash fa-fw' aria-hidden='true'></i></span>";

            return "
            <div class='mb-3 d-inline-block w-100 live_join_class_conn " . (!$isvisible ? 'eye_disabled' : '') . "'>
                <span style='line-height: 36px;'>{$teacher->firstname} {$teacher->lastname}: {$group->name}</span>
                $eyeicon
                <a href='{$linkroot}{$meetingid}' target='_blank' class='btn btn-primary float-right'>Join</a>
            </div>";
        }

        if ($isvisible && $this->userisingroup($group)) {
            return "
            <div class='mb-3 d-inline-block w-100 live_join_class_conn'>
                <span style='line-height: 36px;'>{$teacher->firstname} {$teacher->lastname}: {$group->name}</span>
                <a href='{$linkroot}{$meetingid}' target='_blank' class='btn btn-primary float-right'>Join</a>
            </div>";
        }

        return '';
    }

    /**
     * Checks if the current user has access to manage the meeting.
     *
     * @param object $teacher The teacher object to compare against.
     * @return bool True if the user is the teacher or a site admin, false otherwise.
     */
    private function userhasaccess($teacher) {
        global $USER;
        return $USER->id === $teacher->id || is_siteadmin($USER);
    }

    /**
     * Checks if the current user is a member of the specified group.
     *
     * @param object $group The group object to check membership for.
     * @return bool True if the user is in the group or if the group ID is 0, false otherwise.
     */
    private function userisingroup($group) {
        global $DB, $USER;
        return $group->id == 0 || $DB->record_exists('groups_members', [
                        'userid' => $USER->id,
                        'groupid' => $group->id,
                ]);
    }

    /**
     * Generates the CSS and JavaScript for the Zoom meeting links.
     *
     * @return string HTML string containing style and script tags for meeting link functionality.
     */
    private function generatestyleandscript() {
        return "
        <style>
            .live_join_class_conn { border: 1px solid #717171; padding: 6px; }
            .live_join_class_conn.eye_disabled { border-color: #989898; color: #989898; }
            .live_join_class_eye:not(.eye_disabled) .icon.fa { cursor: pointer; }
        </style>
        <script>
            setTimeout(function() {
                $('.live_join_class_eye').click(function() {
                    var meetingid = $(this).data('id');
                    $.post('/blocks/zoomonline/ajax.php', {
                        id: meetingid,
                        visible: $(this).data('visible')
                    }, function(data) {
                        window.location.reload();
                    });
                });
            }, 4000);
        </script>";
    }

    /**
     * Specifies the course formats where this block can be used.
     *
     * @return array An associative array of course formats and their allowed status.
     */
    public function applicable_formats() {
        return [
                'all' => false,
                'site' => true,
                'site-index' => true,
                'course-view' => true,
                'course-view-social' => false,
                'mod' => true,
                'mod-quiz' => false,
        ];
    }

    /**
     * Indicates whether multiple instances of this block are allowed on a single page.
     *
     * @return bool Always returns true, allowing multiple instances.
     */
    public function instance_allow_multiple() {
        return true;
    }

    /**
     * Indicates whether this block has a configuration page.
     *
     * @return bool Always returns true, indicating this block has a configuration.
     */
    public function has_config() {
        return true;
    }
}
